/* -*-C-*-
*******************************************************************************
*
* File:         hpl_pnm.c
* RCS:          $Header: hpl_pnm.c,v 1.1 95/05/05 11:46:13 hmgr Exp $
* Description:  PNM format handler extension to hplsh.  This uses
*               calls from the public domain PBMPlus libraries,
*               so support should be fairly complete.
* Author:       Andrew Fitzhugh
* Created:      Tue May  2 18:30:11 1995
* Modified:     Mon May 15 14:06:49 1995 (AEF) fitz@hplaef.hpl.hp.com
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
* (C) Copyright 1995, Hewlett-Packard Laboratories, all rights reserved.
*
*******************************************************************************
*/

#include <tcl.h>
#include <tk.h>
#include <pnm.h>

/*
 * The maximum amount of memory to allocate for data read from the
 * file.  If we need more than this, we do it in pieces.
 */

#define MAX_MEMORY	10000		/* don't allocate > 10KB */

static int		FileMatchPNM _ANSI_ARGS_((FILE *f, char *fileName,
			    char *formatString, int *widthPtr,
			    int *heightPtr));
static int		FileReadPNM  _ANSI_ARGS_((Tcl_Interp *interp,
			    FILE *f, char *fileName, char *formatString,
			    Tk_PhotoHandle imageHandle, int destX, int destY,
			    int width, int height, int srcX, int srcY));
static int		FileWritePNM _ANSI_ARGS_((Tcl_Interp *interp,
			    char *fileName, char *formatString,
			    Tk_PhotoImageBlock *blockPtr));

Tk_PhotoImageFormat tkImgFmtPNM = {
    "PNM",			/* name */
    FileMatchPNM,		/* fileMatchProc */
    NULL,			/* stringMatchProc */
    FileReadPNM,		/* fileReadProc */
    NULL,			/* stringReadProc */
    FileWritePNM,		/* fileWriteProc */
    NULL,			/* stringWriteProc */
};

/*
 * Prototypes for local procedures defined in this file:
 */

static int		ReadPNMHeader _ANSI_ARGS_((FILE *f, int *widthPtr,
			    int *heightPtr, int *maxval, int *format));
static void		Convert16to8 _ANSI_ARGS_((gray *pnmRowBuf, 
			    unsigned char *pixelPtr, int fileWidth));
						  


/*
 *----------------------------------------------------------------------
 *
 * FileMatchPNM --
 *
 *	This procedure is invoked by the photo image type to see if
 *	a file contains image data in PNM format.
 *
 * Results:
 *	The return value is 1 if the first characters in file "f" look
 *	like PNM data, and 0 otherwise..
 *
 * Side effects:
 *	The access position in f may change;  if the return value is
 *      TRUE, the access position will be left at the start of the image
 *      data.
 *
 *----------------------------------------------------------------------
 */

static int
FileMatchPNM(f, fileName, formatString, widthPtr, heightPtr)
    FILE *f;			/* The image file, open for reading. */
    char *fileName;		/* The name of the image file. */
    char *formatString;		/* User-specified format string, or NULL. */
    int *widthPtr, *heightPtr;	/* The dimensions of the image are
				 * returned here if the file is a valid
				 * raw PNM file. */
{
    int pnm_format, pnm_maxval;
    int retval;

    retval = ReadPNMHeader(f, widthPtr, heightPtr, &pnm_maxval, &pnm_format);
    return(retval);
}



/*
 *----------------------------------------------------------------------
 *
 * convert16to8 --
 *
 *	This procedure converts pnm data that is 16 bits per component
 *      to 8 bits per component.  This is necessary when linked to PNM
 *      libraries built with the constant PGM_BIGGRAYS defined.
 *
 * Results:
 *	No return value.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */
static void Convert16to8(pnmRowBuf, pixelPtr, fileWidth)
gray *pnmRowBuf;
unsigned char *pixelPtr;
int fileWidth;
{
    int i;
    
    for (i=0; i<fileWidth*3; i++) {
	*pixelPtr++ = (unsigned char) *pnmRowBuf++;
    }
}


/*
 *----------------------------------------------------------------------
 *
 * FileReadPNM --
 *
 *	This procedure is called by the photo image type to read
 *	PNM format data from a file and write it into a given
 *	photo image.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *
 *	The PBMPlus routines are not graceful about returning error
 *      codes to the calling routines;  if there are read errors, the 
 *      PBMPlus routines *might* handle them gracefully.
 *
 * Side effects:
 *	The access position in file f is changed, and new data is
 *	added to the image given by imageHandle.
 *
 *      PBM and PGM data are promoted to PPM in all cases, for simplicity.
 *
 *----------------------------------------------------------------------
 */

static int
FileReadPNM(interp, f, fileName, formatString, imageHandle, destX, destY,
	width, height, srcX, srcY)
    Tcl_Interp *interp;		/* Interpreter to use for reporting errors. */
    FILE *f;			/* The image file, open for reading. */
    char *fileName;		/* The name of the image file. */
    char *formatString;		/* User-specified format string, or NULL. */
    Tk_PhotoHandle imageHandle;	/* The photo image to write into. */
    int destX, destY;		/* Coordinates of top-left pixel in
				 * photo image to be written to. */
    int width, height;		/* Dimensions of block of photo image to
				 * be written to. */
    int srcX, srcY;		/* Coordinates of top-left pixel to be used
				 * in image being read. */
{
    int fileWidth, fileHeight;
    int nLines, nBytes, h, i;
    unsigned char *pixelPtr;
    Tk_PhotoImageBlock block;

    /* PBMPlus variables */
    xel *pnmRowBuf;
    xelval pnmMaxval;
    int pnmFormat;

    if (!ReadPNMHeader(f, &fileWidth, &fileHeight, (int *)&pnmMaxval, 
		       &pnmFormat)) {
	Tcl_AppendResult(interp, "couldn't read raw PNM header from file \"",
		fileName, "\"", NULL);
	return TCL_ERROR;
    }
    if ((fileWidth <= 0) || (fileHeight <= 0)) {
	Tcl_AppendResult(interp, "PNM image file \"", fileName,
		"\" has dimension(s) <= 0", (char *) NULL);
	return TCL_ERROR;
    }

    if ((srcX + width) > fileWidth) {
	width = fileWidth - srcX;
    }
    if ((srcY + height) > fileHeight) {
	height = fileHeight - srcY;
    }
    if ((width <= 0) || (height <= 0)
	|| (srcX >= fileWidth) || (srcY >= fileHeight)) {
	return TCL_OK;
    }

    block.width = width;
    block.pixelSize = 3;
    block.pitch = 3 * fileWidth;
    block.offset[0] = 0;
    block.offset[1] = 1;
    block.offset[2] = 2;

    Tk_PhotoExpand(imageHandle, destX + width, destY + height);

    if (srcY > 0) {
	fseek(f, (long) (srcY * block.pitch), SEEK_CUR);
    }

    nLines = (MAX_MEMORY + block.pitch - 1) / block.pitch;
    if (nLines > height) {
	nLines = height;
    }
    if (nLines <= 0) {
	nLines = 1;
    }
    nBytes = nLines * block.pitch;
    pixelPtr = (unsigned char *) ckalloc((unsigned) nBytes);
    pnmRowBuf = pnm_allocrow(fileWidth);
    block.pixelPtr = pixelPtr + srcX * 3;


    for (h = height; h > 0; h -= nLines) {
	if (nLines > h) {
	    nLines = h;
	    nBytes = nLines * block.pitch;
	}

	for (i=0; i<nLines; i++) {
	    pnm_readpnmrow(f, pnmRowBuf, fileWidth, pnmMaxval, pnmFormat);
	    /* Promoting PBM files does not seem to scale value properly? */
	    if (PNM_FORMAT_TYPE(pnmFormat) == PBM_TYPE) {
		pnm_promoteformatrow(pnmRowBuf, fileWidth, 1, 
				     pnmFormat, (xelval) 255, PPM_TYPE);
	    }
	    else {
		pnm_promoteformatrow(pnmRowBuf, fileWidth, pnmMaxval,
				     pnmFormat, pnmMaxval, PPM_TYPE);
	    }

#ifdef PGM_BIGGRAYS
	    /* Internal PBM gray values are 16 bits instead of 8 */
	    Convert16to8((gray *) pnmRowBuf, pixelPtr+i*fileWidth*3, fileWidth);
#else
	    memcpy(pixelPtr+i*fileWidth*3, (char *) pnmRowBuf, fileWidth*3);
#endif
	}

	block.height = nLines;
	Tk_PhotoPutBlock(imageHandle, &block, destX, destY, width, nLines);
	destY += nLines;
    }

    pnm_freerow(pnmRowBuf);
    ckfree((char *) pixelPtr);
    return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * FileWritePNM --
 *
 *	This procedure is invoked to write image data to a file in PNM
 *	format.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *
 * Side effects:
 *	Data is written to the file given by "fileName".
 *
 *----------------------------------------------------------------------
 */

static int
FileWritePNM(interp, fileName, formatString, blockPtr)
    Tcl_Interp *interp;
    char *fileName;
    char *formatString;
    Tk_PhotoImageBlock *blockPtr;
{
    FILE *f;
    int w, h;
    int greenOffset, blueOffset, nBytes;
    unsigned char *pixelPtr, *pixLinePtr;

    if ((f = fopen(fileName, "wb")) == NULL) {
	Tcl_AppendResult(interp, fileName, ": ", Tcl_PosixError(interp),
		(char *)NULL);
	return TCL_ERROR;
    }

    fprintf(f, "P6\n%d %d\n255\n", blockPtr->width, blockPtr->height);

    pixLinePtr = blockPtr->pixelPtr + blockPtr->offset[0];
    greenOffset = blockPtr->offset[1] - blockPtr->offset[0];
    blueOffset = blockPtr->offset[2] - blockPtr->offset[0];

    if ((greenOffset == 1) && (blueOffset == 2) && (blockPtr->pixelSize == 3)
	    && (blockPtr->pitch == (blockPtr->width * 3))) {
	nBytes = blockPtr->height * blockPtr->pitch;
	if (fwrite(pixLinePtr, 1, (unsigned) nBytes, f) != nBytes) {
	    goto writeerror;
	}
    } else {
	for (h = blockPtr->height; h > 0; h--) {
	    pixelPtr = pixLinePtr;
	    for (w = blockPtr->width; w > 0; w--) {
		if ((putc(pixelPtr[0], f) == EOF)
			|| (putc(pixelPtr[greenOffset], f) == EOF)
			|| (putc(pixelPtr[blueOffset], f) == EOF)) {
		    goto writeerror;
		}
		pixelPtr += blockPtr->pixelSize;
	    }
	    pixLinePtr += blockPtr->pitch;
	}
    }

    if (fclose(f) == 0) {
	return TCL_OK;
    }
    f = NULL;

 writeerror:
    Tcl_AppendResult(interp, "error writing \"", fileName, "\": ",
	    Tcl_PosixError(interp), (char *) NULL);
    if (f != NULL) {
	fclose(f);
    }
    return TCL_ERROR;
}


/*
 *----------------------------------------------------------------------
 *
 * ReadPNMHeader --
 *
 *	This procedure reads the PNM header from the beginning of a
 *	PNM file and returns the dimensions of the image.
 *
 * Results:
 *	The return value is 1 if file "f" appears to start with
 *	a valid PNM header, 0 otherwise.  If the header is valid,
 *	then *widthPtr and *heightPtr are modified to hold the
 *	dimensions of the image.
 *
 * Side effects:
 *	The access position in f advances.
 *
 *----------------------------------------------------------------------
 */

static int
ReadPNMHeader(f, widthPtr, heightPtr, maxval, format)
    FILE *f;			/* Image file to read the header from */
    int *widthPtr, *heightPtr;	/* The dimensions of the image are
				 * returned here. */
    int *maxval;                /* PNM maxval */
    int *format;                /* PNM format code */
{
    int success = 0;


    pnm_readpnminit(f, widthPtr, heightPtr, (xelval *) maxval, format);
  
    switch (PNM_FORMAT_TYPE(*format)) { 
    case PBM_TYPE:
	*maxval = 1;
    case PGM_TYPE:
    case PPM_TYPE:
	success = 1;
        break;
    }
    
    return(success);
}



/*
 *----------------------------------------------------------------------
 *
 * Pnm_Init --
 *
 *	This procedure is typically invoked by Tcl_AppInit procedures
 *	to perform additional Tiff initialization.
 *
 * Results:
 *	Returns a standard Tcl completion code and sets interp->result
 *	if there is an error.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
Pnm_Init(interp)
    Tcl_Interp *interp;		/* Interpreter to initialize. */
{
    Tk_CreatePhotoImageFormat(&tkImgFmtPNM);

    return TCL_OK;
}
